-- imports for Parsec as an applicative
import Control.Applicative
import Control.Monad (MonadPlus(..), ap)
import Text.ParserCombinators.Parsec hiding (many, optional, (<|>), token)

-- the whole program is a Unix filter.
getRight :: Show l => Either l r -> r
getRight (Left  l) = error ("Parse Error:\n" ++ show l)
getRight (Right r) = r

main :: IO ()
main = putStr . getRight . runParser mainParser () "stdin" =<< getContents

-- The parser I use for the whole program
mainParser :: Parser String
mainParser = toplevel

-- some basic parsers
ident :: Parser String
ident = (:) <$>      (letter   <|> char '_')
            <*> many (alphaNum <|> char '_')
        <?> "identifier"

strChar :: Parser Char
strChar = char '\\' *> choice [ char 'n' *> pure '\n'
                              , char 'r' *> pure '\r'
                              , char 't' *> pure '\t'
                              , anyToken ]
          <|> anyToken
          <?> "character embedded in a string"

str :: Parser String -> Parser String
str c = between c c (many (notFollowedBy c *> strChar))

rawStrChar :: Parser String
rawStrChar = do escape <- char '\\' *> pure "\\" <|> pure ""
                c      <- anyToken
                pure $ escape ++ [c]
             <?> "character embedded in a string"

rawStr :: Parser String -> Parser String
rawStr c = do begin <- c
              chars <- concat <$> many (notFollowedBy c *> rawStrChar)
              end   <- c
              pure $ begin ++ chars ++ end

string1    :: Parser String
string2    :: Parser String
rawString1 :: Parser String
rawString2 :: Parser String
string1    = str    (try (string "\'")) <?> "string1"
string2    = str    (try (string "\"")) <?> "string2"
rawString1 = rawStr (try (string "\'")) <?> "rawString1"
rawString2 = rawStr (try (string "\"")) <?> "rawString2"

removeBackquotes                     = drop 2 . init . init
unescapeBackquotes []                = []
unescapeBackquotes ('\\' : '`' : xs) = '`' : unescapeBackquotes xs
unescapeBackquotes (x          : xs) = x   : unescapeBackquotes xs
escapeDBQuotes     []                = []
escapeDBQuotes     ('\\' : c : xs)   = '\\' : c   : escapeDBQuotes xs
escapeDBQuotes     ('"'      : xs)   = '\\' : '"' : escapeDBQuotes xs
escapeDBQuotes     (x        : xs)   = x          : escapeDBQuotes xs
toDBQotes = (++ "\"") . ("\"" ++) .
            unescapeBackquotes . escapeDBQuotes . removeBackquotes

newString :: Parser String
newString  = toDBQotes <$> rawStr (try (string "``"))
             <?> "newString"

token :: String -> Parser String
token s = try (spaces *> string s)

lineComment :: Parser String
lineComment = (++) <$> try (string "--") <*> many (noneOf "\n\r")

guard :: Bool -> Parser ()
guard b = if b
          then fail "guard not satisfied"
          else pure ()

-- the main stuff (it's all Parser String)
toplevel = concat <$> many (choice [ rule
                                   , rawString1
                                   , rawString2
                                   , lineComment
                                   , (: []) <$> anyToken ])

nest :: Parser String -> Parser String -> Parser String
nest o c = do begin  <- o
              inside <- nested o c
              end    <- c
              pure $ begin ++ inside  ++ end

nested :: Parser String -> Parser String -> Parser String
nested o c = concat <$> many (choice [ nest o c
                                     , rule
                                     , (: []) <$>
                                       (notFollowedBy (o <|> c) *> anyToken) ])

hostAtom = spaces *>
           choice [ rule
                  , nest (string "(") (string ")")
                  , nest (string "{") (string "}")
                  , nest (string "[") (string "]")
                  , ident
                  , rawString1
                  , rawString2 ]

rule = between (token "{|") (token "|}") pChoice
       <?> "rule"

pChoice = do b <- bind
             choice [ do try (token "|" <* notFollowedBy (char '}')) -- hack
                         c <- pChoice
                         pure $ "(" ++ b ++ "<|>" ++ c ++ ")"
                    , pure b ]
          <?> "pChoice"

bind = do p <- postfix
          choice [ do token ":"
                      i <- ident
                      b <- bind
                      pure $ "(" ++ p ++ " >>= \\" ++ i ++ " -> " ++ b ++ ")"
                 , do b <- bind
                      pure $ "(" ++ p ++ " >> " ++ b ++ ")"
                 , pure p ]
       <?> "bind"

iter :: String -> Parser String
iter a = choice [ char '*' *> pure ("many (" ++ a ++ ")") -- can't use '$' ???
                , pure a ]
         <* spaces

call :: String -> Parser String
call a = choice [ do args <- tail . init <$> nest (string "(") (string ")")
                     call ("(" ++ a ++ " " ++ args ++ ")")
                , spaces *> pure a ]

postfix = do a <- atom
             b <- call a
             iter b
          <?> "postfix"

atom = spaces *>
       choice
       [ token "?"  *> hostAtom >>= \a -> pure $ "(guard "         ++ a ++ ")"
       , token "->" *> hostAtom >>= \a -> pure $ "(pure "          ++ a ++ ")"
       , token "~"  *> postfix  >>= \a -> pure $ "(notFollowedBy " ++ a ++ ")"
       , token "_"              >>        pure   "spaces"
       , ident
       , rawString1             >>= \s -> pure $ "(char "          ++ s ++ ")"
       , rawString2             >>= \s -> pure $ "(token "         ++ s ++ ")"
       , newString              >>= \s -> pure $ "(try (string "   ++ s ++ "))"
       , between (token "(") (token ")") pChoice
       ]
       <?> "atom"
