-- imports for Parsec as an applicative
import Control.Applicative
import Control.Monad (MonadPlus(..), ap)
import Text.ParserCombinators.Parsec hiding (many, optional, (<|>), token)

-- the whole program is a Unix filter.
getRight :: Show l => Either l r -> r
getRight (Left  l) = error ("Parse Error:\n" ++ show l)
getRight (Right r) = r

main :: IO ()
main = putStr . getRight . runParser mainParser () "stdin" =<< getContents

-- The parser I use for the whole program
mainParser :: Parser String
mainParser = toplevel

-- some basic parsers
ident :: Parser String
ident = {| (letter   | '_')  :x
           (alphaNum | '_')* :xs -> (x:xs) |}
        <?> "identifier"

strChar :: Parser Char
strChar = {| '\\' ( 'n' -> '\n'
                  | 'r' -> '\r'
                  | 't' -> '\t'
                  | anyToken)
           | anyToken |}
          <?> "character embedded in a string"

str :: Parser String -> Parser String
str c = {| c (~c strChar)*:s c -> s |}

rawStrChar :: Parser String
rawStrChar = {| ('\\' -> "\\" | -> ""):escape anyToken:c -> (escape ++ [c]) |}
             <?> "character embedded in a string"

rawStr :: Parser String -> Parser String
rawStr c = {| c:b (~c rawStrChar)*:s c:e -> (b ++ concat s ++ e) |}

string1    :: Parser String
string2    :: Parser String
rawString1 :: Parser String
rawString2 :: Parser String
string1    = str    {| ``\'`` |} <?> "string1"
string2    = str    {| ``\"`` |} <?> "string2"
rawString1 = rawStr {| ``\'`` |} <?> "rawString1"
rawString2 = rawStr {| ``\"`` |} <?> "rawString2"

removeBackquotes                     = drop 2 . init . init
unescapeBackquotes []                = []
unescapeBackquotes ('\\' : '`' : xs) = '`' : unescapeBackquotes xs
unescapeBackquotes (x          : xs) = x   : unescapeBackquotes xs
escapeDBQuotes     []                = []
escapeDBQuotes     ('\\' : c : xs)   = '\\' : c   : escapeDBQuotes xs
escapeDBQuotes     ('"'      : xs)   = '\\' : '"' : escapeDBQuotes xs
escapeDBQuotes     (x        : xs)   = x          : escapeDBQuotes xs
toDBQotes = (++ "\"") . ("\"" ++) .
            unescapeBackquotes . escapeDBQuotes . removeBackquotes

newString :: Parser String
newString  = toDBQotes <$> rawStr (try (string "``"))
             <?> "newString"

token :: String -> Parser String
token s = try {| _ string(s) |}

lineComment :: Parser String
lineComment = {| ``--`` noneOf("\n\r")*:l -> ("--" ++ l) |}

guard :: Bool -> Parser ()
guard b = if b
          then fail "guard not satisfied"
          else pure ()

-- the main stuff (it's all Parser String)
toplevel = {| ( rule | rawString1 | rawString2 | lineComment | anyToken:t -> [t]
              )*:l -> (concat l) |}

nest :: Parser String -> Parser String -> Parser String
nest o c = {| o:begin nested(o c):inside c:end -> (begin ++ inside  ++ end) |}

nested :: Parser String -> Parser String -> Parser String
nested o c = {| (nest(o c) | rule | ~(o | c) anyToken:t -> [t]
                )*:xs -> (concat xs) |}

hostAtom = {| _ ( rule
                | nest((string "(") (string ")"))
                | nest((string "{") (string "}"))
                | nest((string "[") (string "]"))
                | ident
                | rawString1
                | rawString2) |}

rule = {| "{|" pChoice:c  "|}" -> c |}
       <?> "rule"

pChoice = {| bind:b ( try({| "|" ~'}' |}) pChoice:c -> ("(" ++ b ++
                                                            "<|>" ++ c ++ ")")
                    |                               -> b) |}
          <?> "pChoice"

bind = {| postfix:p ( ":" ident:i bind:b -> ("(" ++ p ++ " >>= \\"
                                                 ++ i ++ " -> " ++ b ++ ")")
                    |             bind:b -> ("(" ++ p ++ " >> " ++ b ++ ")")
                    |                    ->         p) |}
       <?> "bind"

iter :: String -> Parser String
iter a = {| ('*'  -> ("many (" ++ a ++ ")")
            |     ->              a) |}
         <* spaces

call :: String -> Parser String
call a = {| nest((string "(") (string ")")):args
            call(("(" ++ a ++ " " ++ (tail . init $ args) ++ ")"))
          | _ -> a |}

postfix = {| atom:a call(a):b iter(b) |}
          <?> "postfix"

atom = {| _ ( "?"  hostAtom:a   -> ("(guard "         ++ a ++ ")")
            | "->" hostAtom:a   -> ("(pure "          ++ a ++ ")")
            | "~"  postfix :a   -> ("(notFollowedBy " ++ a ++ ")")
            | "_"               -> "spaces"
            | ident
            | rawString1   :s   -> ("(char "          ++ s ++ ")")
            | rawString2   :s   -> ("(token "         ++ s ++ ")")
            | newString    :s   -> ("(try (string "   ++ s ++"))")
            | "(" pChoice:c ")" -> c) |}
       <?> "atom"
